/*
 * testap.java
 *
 * Created on November 11, 2002, 3:31 PM
 */

package demo;

import Secret_Share.*;
import java.math.BigInteger;
import java.security.SecureRandom;

/** This is the initial test application. It performs a test of all of the available functionality.
 *
 * The two arguments for the main function are n and k. "n" is the number of clients and "k" is the number of clients required for recovery.
 * @author Brice Daniels
 * @version .3
 */
public class testap {
    
    /** Creates a new instance of testap */
    public testap() {
    }
    
    /** <P>The number of bits, used as a constant throughout {@link testap testap}.</P>
     */    
    public static int numBits = 128;
    
    /** Executes a test of all of the routines with the arguments n and k.
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        //check the arguments
        //args[0] = n
        //args[1] = k
        if(args.length != 2) {
            System.out.println("Wrong number of args");
            System.exit(-1);
        }
        
        int n = 0;
        int k = 0;
        
        try {
            n = Integer.parseInt(args[0]);
            k = Integer.parseInt(args[1]);
        }
        catch (NumberFormatException e) {
            System.out.println("Both arguments must be numbers");
            System.exit(-1);
        }
        
        //create the required Secure Random source
        SecureRandom sr = new SecureRandom();
        
        //generate p, q, g and h
        BigInteger[] global_values = ShareUtil.generateValues(sr,testap.numBits);
        
        //print p, q, g and h
        System.out.println("Global Variables:");
        System.out.println("\tp - " + global_values[0].toString());
        System.out.println("\tq - " + global_values[1].toString());
        System.out.println("\tg - " + global_values[2].toString());
        System.out.println("\th - " + global_values[3].toString());
        
        //randomly generate the secret
        BigInteger secret = new BigInteger(testap.numBits, sr);
        
        //create the E array
        BigInteger[] eArray = new BigInteger[k];
        
        //share the secret
        Share[] shares = null;
        try {
            shares = ShareUtil.split(secret,n,k,testap.numBits,global_values[0],global_values[1],global_values[2],global_values[3],eArray);
        } catch (SplitException e) {
            System.out.println("Error splitting the shares - " + e.getMessage());
            System.exit(-1);
        } catch (ShareUtilException e) {
            System.out.println("Error splitting the shares - " + e.getMessage());
            System.exit(-1);
        }
        
        int i;
        int j;
        
        //print the eArray
/*        System.out.print("E Array: ");
        for(i = 0;i < eArray.length;i++) {
            System.out.print(eArray[i].toString());
            if(i != eArray.length - 1) {
                System.out.print(", ");
            }
        }
        System.out.println();
 */
        
        //verify the shares
        for(i = 0;i < n;i++) {
            try {
//                System.out.println("Share - " + shares[i].getShareNumber() + ": " + shares[i].getSecretShare() + " - " + shares[i].getSecretCommitment());
                if(!shares[i].verify(eArray)) {
                    System.out.println("Share " + (i+1) + " is not valid against the eArray");
                } else {
                    System.out.println("Share " + (i+1) + " is valid against the eArray");
                }
            } catch (SecretShareException e) {
                System.out.println("Share " + (i+1) + " encountered an error during verification - "
                + e.getMessage());
                System.exit(-1);
            }
        }
        
        //renew the secret shares
        System.out.println("Beginning share renewal");
        Share[][] renewalShares = new Share[n][];
        BigInteger[][] renewalEArray = new BigInteger[n][];
        for(i = 0;i < n;i++) {
            renewalEArray[i] = new BigInteger[k];
            try {
                renewalShares[i] = ShareUtil.split(BigInteger.ZERO,n,k,testap.numBits,global_values[0],global_values[1],global_values[2],global_values[3],renewalEArray[i]);
            } catch (SplitException e) {
                System.out.println("Split error during zero split " + (i+1) + " " + e.getMessage());
                System.exit(-1);
            } catch (ShareUtilException e) {
                System.out.println("Split error during zero split " + (i+1) + " " + e.getMessage());
                System.exit(-1);
            }
        }
        for(i = 0;i < n;i++) {
            for(j = 0;j < n;j++) {
                //for each subshare, it must be verified in three ways
                //1. verify the signature of the message
                //2. verify the share given the public information supplied
                //3. verify that the share uses the same basis as the original share
                //steps 2 and 3 are performed by Share.add().
                //in a real situation, it would be necessary to do 1 and 2 before executing Share.add()
                try {
                    shares[i].add(renewalShares[j][i],renewalEArray[j]);
                } catch (ShareAddException e) {
                    //There is a serious problem with the share - accuse client j
                    System.out.println("Subshare " + (j+1) + ": " + e.getMessage());
                    System.exit(-1);
                } catch (ShareUtilException e) {
                    System.out.println("System error during addition for subshare " + (j+1) + ": " + e.getMessage());
                    System.exit(-1);
                } catch (ShareException e) {
                    System.out.println("Invalid commitment value length");
                    System.exit(-1);
                }
                try {
                    if(!shares[i].selfVerify()) {
                        System.out.println("Error selfVerify of share " + (i+1) + " after adding " + (j+1));
                        System.exit(-1);
                    } else {
//                        System.out.println("Successful selfVerify of share " + (i+1) + " after adding " + (j+1));
                    }
                } catch (ShareUtilException e) {
                    System.out.println("Error attempting to selfVerify share " + (i+1));
                    System.exit(-1);
                }
            }
        }
        System.out.println("Finished share renewal");
        
        //recover a share
        //make share n/2 bad (round down)
        int badShare = n/2;
        BigInteger badShareNumber = new BigInteger(Integer.toString(badShare));
        //display the original share
        System.out.println("Bad share used to be - " + shares[badShare - 1].getSecretShare().toString());
        shares[badShare - 1].corruptShare(testap.numBits);
        System.out.println("Bad share now is - " + shares[badShare - 1].getSecretShare().toString());
        System.out.println("Bad share - " + badShare);
        BigInteger[][] recoveryEArray = new BigInteger[n][];
        Share[][] recoverySubShares = new Share[n][];
        for(i = 0;i < n;i++) {
            try {
                if(!shares[i].selfVerify()) {
                    System.out.println("Share " + (i+1) + " fail self-verify check");
                    continue;
                }
            } catch (ShareUtilException e) {
                System.out.println("Error attempting to self-verify share " + (i+1));
                System.exit(-1);
            }
            recoveryEArray[i] = new BigInteger[k];
            try {
                recoverySubShares[i] = ShareUtil.split(BigInteger.ZERO,n,k,testap.numBits,global_values[0],global_values[1],global_values[2],global_values[3],recoveryEArray[i],badShareNumber);
            } catch (SplitException e) {
                System.out.println("Variable Error creating the recovery subshares from share " + (i+1));
                System.exit(-1);
            } catch (ShareUtilException e) {
                System.out.println("Error creating the recovery subshares from share " + (i+1));
                System.exit(-1);
            }
        }
        //every sub share's e array should be verified that it is a share of f(r) = 0
        Share recoverTest = null;
        try {
            recoverTest = new Share(badShare,BigInteger.ZERO,BigInteger.ZERO,global_values[0],global_values[1],global_values[2],global_values[3],recoverySubShares[0][0].getAllPublicVersion(),k);
        } catch (ShareException e) {
            System.out.println("Error creating recovery test share");
            System.exit(-1);
        }
        Share[] recoveryShares = new Share[n];
        for(i = 0;i < n;i++) {
            try {
                if(!shares[i].selfVerify()) {
                    continue;
                }
            } catch (ShareUtilException e) {
                System.out.println("Error attempting to self-verify share " + (i+1));
                System.exit(-1);
            }
            try {
                if(!recoverTest.verify(recoveryEArray[i])) {
                    System.out.println("Recover shares generated by " + (i+1) + " failed recover test");
                    continue;
                }
            } catch (SecretShareException e) {
                System.out.println("Error verifying recoveryEArray " + (i+1));
                System.exit(-1);
            }
            try {
                recoveryShares[i] = shares[i].copy();
            } catch (ShareException e) {
                System.out.println("Error copying share " + (i+1));
                System.exit(-1);
            }
            for(j = 0;j < n;j++) {
                try {
                    if(!shares[j].selfVerify()) {
                        continue;
                    }
                } catch (ShareUtilException e) {
                    System.out.println("Error attempting to self-verify share " + (i+1));
                    System.exit(-1);
                }
                try {
                    if(!recoverySubShares[j][i].verify(recoveryEArray[j])) {
                        System.out.println("Recover subshare " + (j+1) + ", " + (i+1) + " failed to verify");
                        continue;
                    }
                } catch (SecretShareException e) {
                    System.out.println("Error attempting to verify subshare " + (j+1) + ", " + (i+1));
                    System.exit(-1);
                }
                try {
                    recoveryShares[i].add(recoverySubShares[j][i],recoveryEArray[j]);
                } catch (ShareAddException e) {
                    System.out.println("Error attempting to add subshare " + (j+1) + ", " + (i+1) + " to recovery share " + (i+1));
                    System.exit(-1);
                } catch (ShareUtilException e) {
                    System.out.println("Share Util error attempting to add subshare " + (j+1) + ", " + (i+1) + " to recovery share " + (i+1));
                    System.exit(-1);
                } catch (ShareException e) {
                    System.out.println("Invalid commitment value array length");
                    System.exit(-1);
                }
            }
        }
        //The bad share must create a public version of each of the recovery shares
        //it accomplishes that task by using the broadcast EArrays, multiplying them all together
        //I'm going to fake it here using a selfVerify() of the recoveryShares
        for(i = 0;i < n;i++) {
            try {
                if(!shares[i].selfVerify()) {
                    continue;
                }
            } catch (ShareUtilException e) {
                System.out.println("Error attempting to self-verify share " + (i+1));
                System.exit(-1);
            }
            try {
                if(!recoveryShares[i].selfVerify()) {
                    System.out.println("Recovery Share " + (i+1) + " failed its self-verify check");
                    continue;
                }
            } catch (ShareUtilException e) {
                System.out.println("Error attempting to self-verify recovery share " + (i+1));
                System.exit(-1);
            }
        }
        BigInteger recoverSecretShare = null;
        BigInteger recoverCommitShare = null;
        try {
            recoverSecretShare = ShareUtil.recover(recoveryShares,badShareNumber,ShareUtil.RECOVER_SECRET);
        } catch (RecoverException e) {
            System.out.println("Error recovering Secret Share for share recovery");
            System.exit(-1);
        } catch (ShareUtilException e) {
            System.out.println("Share Util Error recovering Secret Share for share recovery");
            System.exit(-1);
        } catch (ShareException e) {
            System.out.println("Invalid commitment value array length");
            System.exit(-1);
        }
        try {
            recoverCommitShare = ShareUtil.recover(recoveryShares,badShareNumber,ShareUtil.RECOVER_COMMIT);
        }  catch (RecoverException e) {
            System.out.println("Error recovering Commitment Share for share recovery");
            System.exit(-1);
        } catch (ShareUtilException e) {
            System.out.println("Share Util Error recovering Commitment Share for share recovery");
            System.exit(-1);
        } catch (ShareException e) {
            System.out.println("Invalid commitment value array length");
            System.exit(-1);
        }
        try {
            shares[badShare - 1] = new Share(badShare,recoverSecretShare,recoverCommitShare,shares[0].getP(),shares[0].getQ(),shares[0].getG(),shares[0].getH(),shares[0].getAllPublicVersion(),shares[0].getK());
        } catch (ShareException e) {
            System.out.println("Error creating recovered share");
            System.exit(-1);
        }
        try {
            if(shares[badShare - 1].selfVerify()) {
                System.out.println("Bad share " + (badShare) + " successfully recovered - " + shares[badShare - 1].getSecretShare().toString());
            }
        } catch (ShareUtilException e) {
            System.out.println("Error processing self-verify check for recovered share");
            System.exit(-1);
        }
        
        //recover the secret
        BigInteger recoveredSecret = null;
        try {
            recoveredSecret = ShareUtil.recover(shares,null,ShareUtil.RECOVER_SECRET);
        } catch (RecoverException e) {
            System.out.println("Error recovering shares - " + e.getMessage());
            System.exit(-1);
        } catch (ShareUtilException e) {
            System.out.println("Error recovering shares - " + e.getMessage());
            System.exit(-1);
        } catch (ShareException e) {
            System.out.println("Invalid commitment value array length");
            System.exit(-1);
        }
        
        //compare recovered secret to original secret
        if(secret.compareTo(recoveredSecret) == 0) {
            System.out.println("Secret correctly recovered");
        } else {
            System.out.println("Secret not correctly recovered");
            System.out.println("Secret: " + secret.toString());
            System.out.println("Recovered Secret: " + recoveredSecret.toString());
        }
    }
}
